// CsToken.h

#ifndef _CS_TOKEN_H_
#define _CS_TOKEN_H_

#pragma warning ( disable : 4311 )
#pragma warning ( disable : 4312 )

#include "stdafx.h"

#include "type.h"
#include "CsTokenTypes.hpp"
#include "HashString.h"


#include <string>

using namespace HashString;

//------------------------------------------------------------------------------
// LineInfo
//------------------------------------------------------------------------------

struct LineInfo
{
	unsigned line : 21;
	unsigned column : 11;

	LineInfo()
		: line(0)
		, column(0)
	{}

	LineInfo ( unsigned line_, unsigned column_ )
		: line ( line_ )
		, column ( column_ )
	{}

	bool operator== ( const LineInfo & rhs ) const
	{
		return line == rhs.line && column == rhs.column;
	}

	bool operator!= ( const LineInfo & rhs ) const
	{
		return line != rhs.line || column != rhs.column;
	}
};

//------------------------------------------------------------------------------
// CsLiteralType
//------------------------------------------------------------------------------

enum CsLiteralType
{
	literal_unknown, // error
	literal_int,
	literal_uint,
	literal_long,
	literal_ulong,
	literal_long_L,
	literal_ulong_L,
	literal_uint_U,
	literal_ulong_U,
	literal_ulong_UL,
	literal_double,
	literal_float_F,
	literal_double_D,
	literal_decimal_M,
	literal_char,
	literal_string,
	literal_verbatim_string,
};

//------------------------------------------------------------------------------
// CsTokenIndex
//------------------------------------------------------------------------------


typedef unsigned CsTokenIndex;


//------------------------------------------------------------------------------
// CsToken
//------------------------------------------------------------------------------

struct CsToken
{
	CsToken()
	: type(tkINVALID)
	, number(0)
	, index(0)
	{}

//	explicit CsToken
//	(
//		CsTokenType token_type
//	)
//	: type(token_type)
//	, line(0)
//	, number(0)
//#if CS_TRANSFORMATION_PARSER
//	, index(0)
//#endif // CS_TRANSFORMATION_PARSER
//	{}

	CsToken
	(
		CsTokenType type_,
		LineInfo line_info_
	)
	: type ( type_ )
	, line_info ( line_info_)
	, number(0)
	, index(0)
	{}

	CsToken
	(
		CsTokenType type_,
		const std::string & text_,
		LineInfo line_info_
	)
	: type(type_)
	, text((const Char*)text_.c_str(),(unsigned)0) // not eval hash string
	, line_info(line_info_)
	, number(0)
	, index(0)
	{}

	CsToken
	(
		CsTokenType type_,
		const Char * begin,
		const Char * end,
		LineInfo line_info_
	)
	: type(type_)
	, text(begin,end,0)  // not eval hash string
	, line_info(line_info_)
	, number(0)
	, index(0)
	{
	}

	CsToken
	(
		CsTokenType type_,
		const Char * begin,
		const Char * end,
		hash_code_type hash_code,
		LineInfo line_info_
	)
	: type(type_)
	, text(begin,end,hash_code)  // not eval hash string
	, line_info(line_info_)
	, number(0)
	, index(0)
	{
	}

	//CsToken
	//(
	//	CsTokenType type_,
	//	const Char * begin,
	//	const Char * end,
	//	unsigned line_, // not eval hash string
	//	bool verbatim
	//)
	//: type(type_)
	//, text(begin,end,0)
	//, line(line_)
	//, number(verbatim)
	//{
	//	ASSERT ( type == tkLITERAL );
	//}

	// for literals
	CsToken
	(
		CsTokenType type_,
		const Char * begin, // not eval hash
		const Char * end,
		LineInfo line_info_,
		CsLiteralType literal_type
	)
	: type(type_)
	, text(begin,end,0)
	, line_info(line_info_)
	, number(literal_type)
	, index(0)
	{
		ASSERT ( type == tkLITERAL || type == tkINVALID );
	}

	CsToken
	(
		CsTokenType type_,
		PHashString phs,
		LineInfo line_info_
	)
	: type(type_)
	, text(phs)
	, line_info(line_info_)
	, number(0)
	, index(0)
	{}

	CsToken
	(
		CsTokenType type_,
		PHashString phs,
		CsLiteralType literal_type,
		LineInfo line_info_
	)
	: type(type_)
	, text(phs)
	, line_info(line_info_)
	, number(literal_type)
	, index(0)
	{
	}

	//CsToken
	//(
	//	CsTokenType type_,
	//	unsigned line_,
	//	unsigned macro_argument_
	//)
	//:	type(type_),
	//	line(line_),
	//	number(macro_argument_)
	//{}

	CsToken
	(
		CsTokenType type_,
		LineEnum line_type, // 0 - \r\n, 1 - \n, 2 - \r
		LineInfo line_info_
	)
	:	type(type_)
	, line_info(line_info_)
	, number((int)line_type)
	, index(0)
	{}

#if CS_PARSER2
#endif // CS_PARSER2

	CsToken
	(
		CsTokenType type_,
		unsigned number_,
		LineInfo line_info_
	)
	:	type(type_)
	, line_info(line_info_)
	, number(number_)
	, index(0)
	{}

  //CsToken
	//(
	//	CsTokenType type_,
	//	FileCache * fileCache
	//)
	//:	type(type_),
	//	number(reinterpret_cast<int>(fileCache))
	//{}

	inline WPHashString getWPHashString() const
	{
		return text;
	}

	//inline PHashString getPHashString() const
	//{
	//	return text;
	//}

	inline const Char * getString() const
	{
		return HashString::getString ( text );
	}

	inline unsigned getStringLength() const
	{
		return getLength ( text );
	}

	const Char * getText() const;

	inline bool isVerbatim() const
	{
//#ifdef _DEBUG
//		ASSERT ( type == tkLITERAL );
//		const Char * s = getString();
//		Char firstChar = s[0];
//		Char lastChar = s[getLength ( text ) - 1];
//		ASSERT ( ( firstChar == '\'' || firstChar == '\"' )
//		      && ( lastChar == '\'' || lastChar == '\"' ) );
//		ASSERT ( number == 0 || number == 1 );
//#endif
		return number == literal_verbatim_string;
	}

	inline bool isLastCharIsDot() const
	{
		const Char * s = getString();
		int length = getStringLength();
		return s[length-1] == '.';
	}

	inline bool operator== ( const CsToken & rhs ) const
	{
		if ( type != rhs.type )
			return false;
		return equal ( text, rhs.text );
	}

	inline bool operator!= ( const CsToken & rhs ) const
	{
		return !operator== ( rhs );
	}

	inline LineInfo getLineInfo() const
	{
		return line_info;
	}

  // return token size in chars
	unsigned evalSize() const;

	unsigned __int8 type;
	unsigned __int8 number;
	PHashString text;
	LineInfo line_info;
	unsigned index;
};

//------------------------------------------------------------------------------

#endif // _CS_TOKEN_H_