// FastDeque.h

#pragma once

#ifndef _FAST_DEQUE_H_
#define _FAST_DEQUE_H_

//#include "allocator_extern.h"
#include "allocator2.h"
#include "allocator_id.h"
#include "type.h"
#include <xutility>


#undef INVARIANT

#define INVARIANT()

template < typename T, typename DequeAllocator, unsigned chunk_size >
class FastDequeImpl;


//#ifdef _DEBUG
//static bool bCheckPushBack = 0;
//#endif // _DEBUG

//--------------------------------------------------------------
// SmallDequeAllocator
//--------------------------------------------------------------

template < typename T, unsigned chunk_size >
class SmallDequeAllocator
{
	template < typename T, typename DequeAllocator, unsigned chunk_size > friend class FastDequeImpl;
private:
	static inline byte * allocate()
	{
		byte * address = (byte*)FixedSmallObjAllocator<sizeof(T)*chunk_size,DefaultAllocatorId,OwnPageAllocator>::allocate();
		return (byte*)address;
	}

	static inline void deallocate ( byte * address )
	{
		FixedSmallObjAllocator<sizeof(T)*chunk_size,DefaultAllocatorId,OwnPageAllocator>::deallocate ( address );
	}
};

//--------------------------------------------------------------
// MediumDequeAllocator
//--------------------------------------------------------------

class MediumDequeAllocator
{
	//friend template < typename T, typename DequeAllocator, unsigned chunk_size >
	//class FastDequeImpl;
public:
//#if CS_PARSER_LIBRARY_USER
//	static const int page_size = 4096;
//#else
	static const int page_size = OwnPageAllocator::page_size;
//#endif // CS_PARSER_LIBRARY_USER

	static inline byte * allocate()
	{
		byte * address = (byte*) OwnPageAllocator::allocate();
		return (byte*)address;
	}

	static inline void deallocate ( byte * address )
	{
		OwnPageAllocator::deallocate ( address );
	}
};

//--------------------------------------------------------------
// FastDequeImpl
//--------------------------------------------------------------

class fast_deque_iterator_tag : public std::forward_iterator_tag
{};

template < typename T, typename DequeAllocator, unsigned chunk_size >
class FastDequeImpl
{
public:
	class forward_iterator
	{
		//friend void _Advance<class _InIt, class _Diff> ( _InIt & it, _Diff distance, fast_deque_iterator_tag );

	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
			typedef fast_deque_iterator_tag iterator_category;
			typedef T value_type;
			typedef ptrdiff_t difference_type;
			typedef ptrdiff_t distance_type;	// retained
			typedef T * pointer;
			typedef T & reference;
			static const unsigned chunk_size = chunk_size;
		//};

		friend FastDequeImpl;
	//private:
	public:
		forward_iterator ( T * current, T * part_end )
			: current_ ( current )
			, part_end_ ( part_end )
		{}

	public:
		forward_iterator()
			: current_ ( 0 )
			, part_end_ ( 0 )
		{}

		T & operator*() const
		{
			return * current_;
		}

		T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const forward_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const forward_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		forward_iterator & operator++ () // preincrement
		{
			++current_;
			if ( current_ == part_end_ )
			{
				//T ** p_part_next = reinterpret_cast<T**>(part_end_);
				T * part_next = * part_next_;
				if ( part_next )
				{
					current_ = part_next+1;
					//part_end_ = part_next+chunk_size_-1;
					part_end_ = part_next+chunk_size_-1;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return *this;
		}

		forward_iterator operator++ ( int ) // postincrement
		{
			forward_iterator retValue = *this;
			++current_;
			if ( current_ == partial_end_ )
			{
				//T ** p_part_next = reinterpret_cast<T**>(part_end_);
				T * part_next = * partial_next_;
				if ( part_next )
				{
					current_ = part_next+1;
					//part_end_ = part_next+chunk_size_-1;
					part_end_ = part_next+chunk_size_-1;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return retValue;
		}

		//forward_iterator & operator-- () // predecrement
		//{
		//	--current_;
		//	return *current_;
		//}

		//forward_iterator operator-- ( int ) // postdecrement
		//{
		//	forward_iterator retValue = *this;
		//	--*this;
		//	return retValue;
		//}

	//private:
	public:
		T * current_;
		union
		{
			T * part_end_;
			T ** part_next_;
		};
	};

	class const_forward_iterator
	{
	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
			typedef fast_deque_iterator_tag iterator_category;
			typedef const T value_type;
			typedef ptrdiff_t difference_type;
			typedef ptrdiff_t distance_type;	// retained
			typedef const T * pointer;
			typedef const T & reference;
			static const unsigned chunk_size = chunk_size;
		//};

		friend FastDequeImpl;
	private:
		const_forward_iterator ( const T * current, const T * part_end )
			: current_ ( current )
			, part_end_ ( part_end )
		{}

	public:
		const_forward_iterator()
			: current_ ( 0 )
			, part_end_ ( 0 )
		{}

		const T & operator*() const
		{
			return * current_;
		}

		const T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const const_forward_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_forward_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_forward_iterator & operator++ () // preincrement
		{
			++current_;
			if ( current_ == part_end_ )
			{
				//T ** p_part_next = reinterpret_cast<T**>(const_cast<T*>(partial_end_));
				const T * part_next = * part_next_;
				if ( part_next )
				{
					current_ = part_next+1;
					part_end_ = part_next+chunk_size_-1;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return *this;
		}

		const_forward_iterator operator++ ( int ) // postincrement
		{
			forward_iterator retValue = *this;
			++current_;
			if ( current_ == part_end_ )
			{
				//const T ** p_part_next = static_cast<const T**>(partial_end_);
				const T * part_next = * part_next_;
				if ( partial_next )
				{
					current_ = part_next+1;
					part_end_ = part_next+chunk_size_-1;
				}
				//else
				//{
				//	current_ = 0;
				//}
			}
			return retValue;
		}

		//forward_iterator & operator-- () // predecrement
		//{
		//	--current_;
		//	return *current_;
		//}

		//forward_iterator operator-- ( int ) // postdecrement
		//{
		//	forward_iterator retValue = *this;
		//	--*this;
		//	return retValue;
		//}

	public:
		const T * current_;
		union
		{
			const T * part_end_;
			const T ** part_next_;
		};
	};

	typedef forward_iterator iterator;
	typedef const_forward_iterator const_iterator;

	explicit FastDequeImpl()
		: first_ ( 0 )
		, end_ ( 0 )
		, last_ ( 0 )
	{}

	void clear()
	{
		T * next = first_;
		while ( next )
		{
			T * last;
			if ( next + chunk_size_ == end_ )
				last = last_;
			else
				last = next + chunk_size_ - 1;
			for ( T * p = next + 1; p != last; ++p )
				p->~T();
			T * partial_end = next + chunk_size_ - 1;
			T ** p_part_next = reinterpret_cast<T**>(partial_end);
			T * part_next = *p_part_next;
			deallocate ( (byte*)next );
			next = part_next;
		}
		first_ = 0;
		end_ = 0;
		last_ = 0;
	}

	~FastDequeImpl()
	{
		clear();
	}

	void resize0()
	{
		T * next = first_;
		while ( next )
		{
			T * last;
			if ( next + chunk_size_ == end_ )
				last = last_;
			else
				last = next + chunk_size_ - 1;
			for ( T * p = next + 1; p != last; ++p )
				p->~T();
			T * partial_end = next + chunk_size_ - 1;
			T ** p_part_next = reinterpret_cast<T**>(partial_end);
			T * part_next = *p_part_next;
			if ( next != first_ )
				deallocate ( (byte*)next );
			next = part_next;
		}
		if ( first_ )
		{
			end_ = first_ + chunk_size_;
			last_ = first_ + 1;
		}
	}

private:
	static byte * allocate()
	{
		return DequeAllocator::allocate();
		//byte * address = (byte*)FixedSmallObjAllocator<sizeof(T)*chunk_size_,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		////byte * address = new byte[sizeof(T)*chunk_size_];
		//return (byte*)address;
	}

	static void deallocate ( byte * address )
	{
		//FixedSmallObjAllocator<sizeof(T)*chunk_size_,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( address );
		////delete address;
		return DequeAllocator::deallocate ( address );
	}

public:
	forward_iterator forward_begin()
	{
		if ( first_ )
			return forward_iterator ( first_ + 1, first_ + chunk_size_ - 1 );
		else
			return forward_iterator ( 0, 0 );
	}

	const_forward_iterator forward_begin() const
	{
		if ( first_ )
			return const_forward_iterator ( first_ + 1, first_ + chunk_size_ - 1 );
		else
			return const_forward_iterator ( 0, 0 );
	}

	forward_iterator forward_end()
	{
		if ( last_ )
			return forward_iterator ( last_, end_ - 1 );
		else
			return forward_iterator ( 0, 0 );
	}

	const_forward_iterator forward_end() const
	{
		if ( last_ )
			return const_forward_iterator ( last_, end_ - 1 );
		else
			return const_forward_iterator ( 0, 0 );
	}

	forward_iterator begin() { return forward_begin(); }
	const_forward_iterator begin() const { return forward_begin(); }
	forward_iterator end() { return forward_end(); }
	const_forward_iterator end() const { return forward_end(); }

	const T & front() const
	{
		return *(first_ + 1);
	}

	T & front()
	{
		return *(first_ + 1);
	}

	const T & back() const
	{
		return *(last_ - 1);
	}

	T & back()
	{
		return *(last_ - 1);
	}

	void push_back ( T value )
	{
		//INVARIANT();

//#ifdef _DEBUG
//		unsigned debug_size_before = eval_size();
//#endif // _DEBUG

		if ( first_ )
		{
			if ( last_ != end_ - 1 )
			{
				new ( last_ ) T ( value );
				++last_;
			}
			else
			{
				T * new_part = reinterpret_cast<T*>(allocate());
				T ** prev_next_part = reinterpret_cast<T**> ( end_ - 1 );
				*prev_next_part = new_part;
				T ** prev_part = reinterpret_cast<T**>(new_part);
				*prev_part = end_ - chunk_size_;
				T ** next_part = reinterpret_cast<T**>(new_part+chunk_size_-1);
				*next_part = 0;
				T * current = new_part + 1;
				new ( current ) T ( value );
				++current;
				end_ = new_part + chunk_size_;
				last_ = current;
			}
		}
		else
		{
			T * new_part = reinterpret_cast<T*>(allocate());
			T ** prev_part = reinterpret_cast<T**>(new_part);
			*prev_part = 0;
			T ** next_part = reinterpret_cast<T**>(new_part+chunk_size_-1);
			*next_part = 0;
			T * current = new_part + 1;
			new ( current ) T ( value );
			++current;
			first_ = new_part;
			end_ = new_part + chunk_size_;
			last_ = current;
		}

//#ifdef _DEBUG
//		unsigned debug_size_after = eval_size();
//		XASSERT ( debug_size_before + 1 == debug_size_after );
//#endif // _DEBUG

		//XASSERT ( last_ <= end_ );
		//XASSERT ( first_ );
		//XASSERT ( last_ );
		//XASSERT ( end_ );

		//INVARIANT();
	}

	template<typename T2>
	void push_back ( T2 it, T2 end )
	{
		for ( ; it != end; ++it )
		{
			push_back ( *it );
		}
	}

	bool empty() const
	{
		return first_ == last_;
	}

	//size_t size() const
	//{
	//	return last_ - first_;
	//}

	//T& operator[] ( unsigned index )
	//{
	//	return first_[index];
	//}

	size_t eval_size() const
	{
		size_t size = 0;
		if ( first_ )
		{
			T * part = first_;
			//XASSERT ( next == prev_part );
			//TRACE1("%p eval_size\n",part);
			while ( part )
			{
				T * next_part = *reinterpret_cast<T**>(&part[chunk_size_-1]);
				if ( next_part )
					size += chunk_size_ - 2;
				else
					size += last_ - ( part + 1 );
				part = next_part;
			}
		}
		return size;
	}

	void setAutoCopy ( FastDequeImpl & rhs )
	{
		first_ = rhs.first_;
		rhs.first_ = 0;
		end_ = rhs.end_;
		rhs.end_ = 0;
		last_ = rhs.last_;
		rhs.last_ = 0;
	}

//#ifdef _DEBUG
//	void invariant() const
//	{
//		if ( first_ )
//		{
//			T * part = first_;
//			T * prev_part = *reinterpret_cast<T**>(&part[0]);
//			XASSERT ( !prev_part );
//			T * next_part = *reinterpret_cast<T**>(&part[chunk_size_-1]);
//			//XASSERT ( next == prev_part );
//			while ( next_part )
//			{
//				T * next_prev_part = *reinterpret_cast<T**>(&next_part[0]);
//				T * next_next_part = *reinterpret_cast<T**>(&next_part[chunk_size_-1]);
//				XASSERT ( part == next_prev_part );
//				part = next_part;
//				prev_part = next_prev_part;
//				next_part = next_next_part;
//			}
//			XASSERT ( !next_part );
//			XASSERT ( ( part + 1 ) <= last_
//				&& last_ <= ( part + chunk_size_ - 1 ) );
//		}
//		else
//		{
//			XASSERT ( first_ == 0 );
//			XASSERT ( end_ == 0 );
//			XASSERT ( last_ == 0 );
//		}
//	}
//#endif // _DEBUG

private:
	static const unsigned chunk_size_ = chunk_size;
	T * first_; // pointer to beginning of array
	T * end_; // pointer to current end of sequence
	T * last_; // pointer to current end of sequence
};

//--------------------------------------------------------------
// MSVC7.1

namespace std {

template<class _InIt, class _Diff> inline //  < FastDequeImpl < CsToken, MediumDequeAllocator, MediumDequeAllocator::page_size/sizeof(CsToken) >, unsigned >
void _Advance ( _InIt & it, _Diff distance, fast_deque_iterator_tag )
{
	typedef _InIt::value_type T;
	typedef _InIt iterator_type;
	union part_ptr_union
	{
		T * part_end_;
		T ** part_next_;
	};

	T * current = it.current_;
	part_ptr_union part_end;
	part_end.part_end_ = it.part_end_;
loop:
	//TRACE1("%p advance\n",current);
	_Diff available = static_cast<_Diff>( part_end.part_end_ - current );
	if ( distance < available )
	{
		it.current_ = current + distance;
		it.part_end_ = part_end.part_end_;
		return;
	}
	T * next_part = *part_end.part_next_;
	if ( next_part )
	{
		distance -= available;
		current = next_part + 1;
		part_end.part_end_ = next_part + _InIt::chunk_size - 1;
		goto loop;
	}
	else
	{
		//XASSERT ( distance == available );
		it.current_ = current + distance;
		it.part_end_ = part_end.part_end_;
		return;
	}
}

} // namespace std

//--------------------------------------------------------------
// FastDeque
//--------------------------------------------------------------

template < typename T >
class FastDeque : public FastDequeImpl<T,SmallDequeAllocator<T,8>,8>
{};

//--------------------------------------------------------------
// MediumFastDeque
//--------------------------------------------------------------

template < typename T >
class MediumFastDeque : public FastDequeImpl<T,MediumDequeAllocator,MediumDequeAllocator::page_size/sizeof(T)>
{};

//--------------------------------------------------------------

//#ifdef _DEBUG
//
//template<>
//void FastDequeImpl < CsToken, MediumDequeAllocator, MediumDequeAllocator::page_size/sizeof(CsToken) >
//::push_back ( CsToken value )
//{
//	if ( bCheckPushBack )
//		XASSERT ( value.type != tkEOF && value.type < 256 );
//	
//	if ( first_ )
//	{
//		if ( last_ != end_ - 1 )
//		{
//			new ( last_ ) CsToken ( value );
//			++last_;
//		}
//		else
//		{
//			CsToken * new_part = reinterpret_cast<CsToken*>(allocate());
//			CsToken ** prev_next_part = reinterpret_cast<CsToken**> ( end_ - 1 );
//			*prev_next_part = new_part;
//			CsToken ** prev_part = reinterpret_cast<CsToken**>(new_part);
//			*prev_part = end_ - chunk_size_;
//			CsToken ** next_part = reinterpret_cast<CsToken**>(new_part+chunk_size_-1);
//			*next_part = 0;
//			CsToken * current = new_part + 1;
//			new ( current ) CsToken ( value );
//			++current;
//			end_ = new_part + chunk_size_;
//			last_ = current;
//		}
//	}
//	else
//	{
//		CsToken * new_part = reinterpret_cast<CsToken*>(allocate());
//		CsToken ** prev_part = reinterpret_cast<CsToken**>(new_part);
//		*prev_part = 0;
//		CsToken ** next_part = reinterpret_cast<CsToken**>(new_part+chunk_size_-1);
//		*next_part = 0;
//		CsToken * current = new_part + 1;
//		new ( current ) CsToken ( value );
//		++current;
//		first_ = new_part;
//		end_ = new_part + chunk_size_;
//		last_ = current;
//	}
//};
//
//#endif // _DEBUG

//--------------------------------------------------------------

#endif // _FAST_DEQUE_H_
